"""Module used to support events and event handling, not to be confused with
the state machine events.
"""

from datetime import datetime
from io import BytesIO
import inspect
import logging
from pathlib import Path
import sys
from typing import Callable, Any, NamedTuple, TYPE_CHECKING, cast, Iterator, Union

from pydicom.dataset import Dataset, FileMetaDataset
from pynetdicom.dimse_primitives import C_STORE
from pydicom.filereader import dcmread
from pydicom.tag import BaseTag
from pydicom.uid import UID

from pynetdicom.dsutils import decode, create_file_meta, encode_file_meta

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.association import Association
    from pynetdicom.dimse_messages import DIMSEMessage
    from pynetdicom.dimse_primitives import (
        C_ECHO,
        C_FIND,
        C_GET,
        C_MOVE,
        N_ACTION,
        N_CREATE,
        N_DELETE,
        N_EVENT_REPORT,
        N_GET,
        N_SET,
    )
    from pynetdicom.pdu import _PDUType
    from pynetdicom.pdu_primitives import SOPClassCommonExtendedNegotiation
    from pynetdicom.presentation import PresentationContextTuple

    _RequestType = (
        C_ECHO
        | C_FIND
        | C_GET
        | C_MOVE
        | C_STORE
        | N_ACTION
        | N_CREATE
        | N_DELETE
        | N_EVENT_REPORT
        | N_GET
        | N_SET
    )


LOGGER = logging.getLogger(__name__)


EventType = Union["NotificationEvent", "InterventionEvent"]
EventHandlerType = tuple[EventType, Callable] | tuple[EventType, Callable, list[Any]]
_BasicReturnType = Dataset | int
_DatasetReturnType = tuple[_BasicReturnType, Dataset | None]
_IteratorType = Iterator[tuple[_BasicReturnType, Dataset | None]]


# Notification events
#   No returns/yields needed, can have multiple handlers per event
class NotificationEvent(NamedTuple):
    """Representation of a notification event.

    .. versionadded:: 1.3

    Possible notification events are:

    * :class:`EVT_ABORTED`
    * :class:`EVT_ACCEPTED`
    * :class:`EVT_ACSE_RECV`
    * :class:`EVT_ACSE_SENT`
    * :class:`EVT_CONN_CLOSE`
    * :class:`EVT_CONN_OPEN`
    * :class:`EVT_DATA_RECV`
    * :class:`EVT_DATA_SENT`
    * :class:`EVT_DIMSE_RECV`
    * :class:`EVT_DIMSE_SENT`
    * :class:`EVT_ESTABLISHED`
    * :class:`EVT_FSM_TRANSITION`
    * :class:`EVT_PDU_RECV`
    * :class:`EVT_PDU_SENT`
    * :class:`EVT_REJECTED`
    * :class:`EVT_RELEASED`
    * :class:`EVT_REQUESTED`
    """

    name: str
    description: str
    is_intervention: bool = False
    is_notification: bool = True

    def __str__(self) -> str:
        """String representation of the class."""
        return self.name


# pylint: disable=line-too-long
EVT_ABORTED = NotificationEvent("EVT_ABORTED", "Association aborted")
EVT_ACCEPTED = NotificationEvent("EVT_ACCEPTED", "Association request accepted")  # noqa
EVT_ACSE_RECV = NotificationEvent(
    "EVT_ACSE_RECV", "ACSE primitive received from DUL"
)  # noqa
EVT_ACSE_SENT = NotificationEvent("EVT_ACSE_SENT", "ACSE primitive sent to DUL")  # noqa
EVT_CONN_CLOSE = NotificationEvent("EVT_CONN_CLOSE", "Connection closed")
EVT_CONN_OPEN = NotificationEvent("EVT_CONN_OPEN", "Connection opened")
EVT_DATA_RECV = NotificationEvent(
    "EVT_DATA_RECV", "PDU data received from remote"
)  # noqa
EVT_DATA_SENT = NotificationEvent("EVT_DATA_SENT", "PDU data sent to remote")
EVT_DIMSE_RECV = NotificationEvent(
    "EVT_DIMSE_RECV", "Complete DIMSE message received and decoded"
)  # noqa
EVT_DIMSE_SENT = NotificationEvent(
    "EVT_DIMSE_SENT", "DIMSE message encoded and P-DATA primitives sent to DUL"
)  # noqa
EVT_ESTABLISHED = NotificationEvent(
    "EVT_ESTABLISHED", "Association established"
)  # noqa
EVT_FSM_TRANSITION = NotificationEvent(
    "EVT_FSM_TRANSITION", "State machine about to transition"
)  # noqa
EVT_PDU_RECV = NotificationEvent("EVT_PDU_RECV", "PDU received and decoded")
EVT_PDU_SENT = NotificationEvent("EVT_PDU_SENT", "PDU encoded and sent")
EVT_REJECTED = NotificationEvent("EVT_REJECTED", "Association request rejected")  # noqa
EVT_RELEASED = NotificationEvent("EVT_RELEASED", "Association released")
EVT_REQUESTED = NotificationEvent("EVT_REQUESTED", "Association requested")


# Intervention events
#   Returns/yields needed if bound, can only have one handler per event
class InterventionEvent(NamedTuple):
    """Representation of an intervention event.

    .. versionadded:: 1.3

    Possible intervention events are:

    * :class:`EVT_ASYNC_OPS`
    * :class:`EVT_SOP_COMMON`
    * :class:`EVT_SOP_EXTENDED`
    * :class:`EVT_USER_ID`
    * :class:`EVT_C_ECHO`
    * :class:`EVT_C_FIND`
    * :class:`EVT_C_GET`
    * :class:`EVT_C_MOVE`
    * :class:`EVT_C_STORE`
    * :class:`EVT_N_ACTION`
    * :class:`EVT_N_CREATE`
    * :class:`EVT_N_DELETE`
    * :class:`EVT_N_EVENT_REPORT`
    * :class:`EVT_N_GET`
    * :class:`EVT_N_SET`
    """

    name: str
    description: str
    is_intervention: bool = True
    is_notification: bool = False

    def __str__(self) -> str:
        """String representation of the class."""
        return self.name


EVT_ASYNC_OPS = InterventionEvent(
    "EVT_ASYNC_OPS", "Asynchronous operations negotiation requested"
)  # noqa
EVT_SOP_COMMON = InterventionEvent(
    "EVT_SOP_COMMON", "SOP class common extended negotiation requested"
)  # noqa
EVT_SOP_EXTENDED = InterventionEvent(
    "EVT_SOP_EXTENDED", "SOP class extended negotiation requested"
)  # noqa
EVT_USER_ID = InterventionEvent(
    "EVT_USER_ID", "User identity negotiation requested"
)  # noqa
EVT_C_ECHO = InterventionEvent("EVT_C_ECHO", "C-ECHO request received")
EVT_C_FIND = InterventionEvent("EVT_C_FIND", "C-FIND request received")
EVT_C_GET = InterventionEvent("EVT_C_GET", "C-GET request received")
EVT_C_MOVE = InterventionEvent("EVT_C_MOVE", "C-MOVE request received")
EVT_C_STORE = InterventionEvent("EVT_C_STORE", "C-STORE request received")
EVT_N_ACTION = InterventionEvent("EVT_N_ACTION", "N-ACTION request received")
EVT_N_CREATE = InterventionEvent("EVT_N_CREATE", "N-CREATE request received")
EVT_N_DELETE = InterventionEvent("EVT_N_DELETE", "N-DELETE request received")
EVT_N_EVENT_REPORT = InterventionEvent(
    "EVT_N_EVENT_REPORT", "N-EVENT-REPORT request received"
)  # noqa
EVT_N_GET = InterventionEvent("EVT_N_GET", "N-GET request received")
EVT_N_SET = InterventionEvent("EVT_N_SET", "N-SET request received")
# pylint: enable=line-too-long

_INTERVENTION_EVENTS = [
    ii[1]
    for ii in inspect.getmembers(
        sys.modules[__name__], lambda x: isinstance(x, InterventionEvent)
    )
]
_NOTIFICATION_EVENTS = [
    ii[1]
    for ii in inspect.getmembers(
        sys.modules[__name__], lambda x: isinstance(x, NotificationEvent)
    )
]


_HandlerBase = tuple[Callable, list[Any] | None]
_NotificationHandlerAttr = list[_HandlerBase]
_InterventionHandlerAttr = _HandlerBase
HandlerArgType = _NotificationHandlerAttr | _InterventionHandlerAttr
_HandlerAttr = dict[EventType, HandlerArgType]


def _add_handler(
    event: EventType, handlers_attr: _HandlerAttr, handler_arg: _HandlerBase
) -> None:
    """Add a handler to an object's handler recording attribute.

    Parameters
    ----------
    event : NotificationEvent or InterventionEvent
        The event the handler should be bound to.
    handlers_attr : dict
        The object attribute of
        {event: [(handler, None | args)] | (handler, None | args)} used to
        record bindings.
    handler_arg : tuple[Callable, None | list[Any]]
        The handler and optional arguments to be bound.
    """
    if isinstance(event, NotificationEvent):
        if event not in handlers_attr:
            handlers_attr[event] = []

        if handler_arg not in handlers_attr[event]:
            h = cast(_NotificationHandlerAttr, handlers_attr[event])
            h.append(handler_arg)

    elif isinstance(event, InterventionEvent):
        # Intervention events - only one handler allowed
        handlers_attr[event] = handler_arg


def _remove_handler(
    event: EventType, handlers_attr: _HandlerAttr, handler: Callable
) -> None:
    """Remove a handler from an object's handler recording attribute.

    Parameters
    ----------
    event : NotificationEvent or InterventionEvent
        The event the handler should be unbound from.
    handlers_attr : dict
        The object attribute of
        {event: list[(handler, None | args)] | (handler, None | args)} used
        to record bindings.
    handler_arg : Callable
        The handler to be unbound.
    """
    if event not in handlers_attr:
        return

    if isinstance(event, NotificationEvent):
        handlers_list = cast(_NotificationHandlerAttr, handlers_attr[event])
        handlers_attr[event] = [h for h in handlers_list if h[0] != handler]
        if not handlers_attr[event]:
            del handlers_attr[event]

    elif isinstance(event, InterventionEvent):
        # Unbind and replace with default
        if handler in handlers_attr[event]:
            handlers_attr[event] = (get_default_handler(event), None)


def get_default_handler(event: InterventionEvent) -> Callable[["Event"], Any]:
    """Return the default handler for an intervention `event`.

    .. versionadded:: 1.3
    """
    handlers = {
        EVT_ASYNC_OPS: _async_ops_handler,
        EVT_SOP_COMMON: _sop_common_handler,
        EVT_SOP_EXTENDED: _sop_extended_handler,
        EVT_USER_ID: _user_identity_handler,
        EVT_C_ECHO: _c_echo_handler,
        EVT_C_FIND: _c_find_handler,
        EVT_C_GET: _c_get_handler,
        EVT_C_MOVE: _c_move_handler,
        EVT_C_STORE: _c_store_handler,
        EVT_N_ACTION: _n_action_handler,
        EVT_N_CREATE: _n_create_handler,
        EVT_N_DELETE: _n_delete_handler,
        EVT_N_EVENT_REPORT: _n_event_report_handler,
        EVT_N_GET: _n_get_handler,
        EVT_N_SET: _n_set_handler,
    }
    return handlers[event]


def trigger(
    assoc: "Association", event: EventType, attrs: dict[str, Any] | None = None
) -> Any | None:
    """Trigger an `event` and call any bound handler(s).

    .. versionadded:: 1.3

    Notification events can be bound to multiple handlers, intervention events
    can only be bound to a single handler.

    **Special Attributes**

    If `attrs` contains:

    * `_is_cancelled` key then :attr:`Event.is_cancelled` will be hooked into
      the value's callable function.
    * a C-FIND, C-GET or C-MOVE `request` key then :attr:`Event.identifier`
      will return the decoded *Identifier* parameter value.
    * a C-STORE `request` key then :attr:`Event.dataset` will return the
      decoded *Data Set* parameter value.
    * an N-ACTION `request` key then :attr:`Event.action_information` will
      return the decoded *Action Information* parameter value.
    * an N-CREATE `request` key then :attr:`Event.attribute_list` will return
      the decoded *Attribute List* parameter value.
    * an N-EVENT-REPORT `request` key then :attr:`Event.event_information` will
      return the decoded *Event Information* parameter value.
    * an N-SET `request` key then :attr:`Event.modification_list` will return
      the decoded *Modification List* parameter value.

    Parameters
    ----------
    assoc : assoc.Association
        The association in which the event occurred.
    event : events.NotificationEvent or events.InterventionEvent
        The event to trigger.
    attrs : dict, optional
        The attributes to set in the :class:`Event` instance that is passed to
        the event's corresponding handler functions as
        ``{attribute name : value}``, default ``{}``.

    Raises
    ------
    Exception
        If an exception occurs in an intervention event handler then the
        exception will be raised. If an exception occurs in a notification
        handler then the exception will be caught and logged instead.
    """
    # Get the handler(s) bound to the event
    #   notification events: returns a list of 2-tuple (callable, args)
    #   intervention events: returns a 2-tuple of (callable, args)
    #       or (None, None)
    handlers = assoc.get_handlers(event)
    # Empty list or (None, None)
    if not handlers or handlers[0] is None:
        return None

    evt = Event(assoc, event, attrs or {})

    try:
        # Intervention event - only single handler allowed
        if isinstance(event, InterventionEvent):
            handlers = cast(_InterventionHandlerAttr, handlers)
            if handlers[1] is not None:
                return handlers[0](evt, *handlers[1])

            return handlers[0](evt)

        # Notification event - multiple handlers are allowed
        handlers = cast(_NotificationHandlerAttr, handlers)
        for func, args in handlers:
            if args:
                func(evt, *args)
            else:
                func(evt)
    except Exception as exc:
        # Intervention exceptions get raised
        if isinstance(event, InterventionEvent):
            raise

        # Capture exceptions for notification events
        LOGGER.error(
            f"Exception raised in user's 'evt.{event.name}' "
            f"event handler '{func.__name__}'"
        )
        LOGGER.exception(exc)

    return None


class Event:
    """Representation of an event.

    .. versionadded:: 1.3

    .. warning::

       Some of :class:`Event`'s attributes are set dynamically when an event is
       triggered and are available only for a specific event. For example, the
       ``Event.context`` attribute is only available for events such as
       ``evt.EVT_C_ECHO``, ``evt.EVT_C_STORE``, etc. See the
       :ref:`handler documentation<api_events>` for a list of what attributes
       are available for a given event.

    Attributes
    ----------
    assoc : association.Association
        The association in which the event occurred.
    timestamp : datetime.datetime
        The date/time the event was created. Will be slightly before or after
        the actual event that this object represents.
    """

    def __init__(
        self,
        assoc: "Association",
        event: EventType,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        """Create a new Event.

        Parameters
        ----------
        assoc : association.Association
            The association in which the event occurred.
        event : events.NotificationEvent or events.InterventionEvent
            The representation of the event that occurred.
        attrs : dict, optional
            The ``{attribute : value}`` pairs to use to set the
            :class:`Event`'s  attributes.
        """
        self.assoc = assoc
        self._event = event
        self.timestamp = datetime.now()

        # Only decode a dataset when necessary
        self._hash: int | None = None
        self._decoded: Dataset | None = None

        # Define type hints for dynamic attributes
        self.request: "_RequestType"
        self._is_cancelled: Callable[[int], bool]
        self.context: "PresentationContextTuple"
        self.current_state: str
        self.fsm_event: str
        self.next_state: str
        self.action: str
        self.data: bytes
        self.message: DIMSEMessage
        self.pdu: _PDUType

        attrs = attrs or {}
        for kk, vv in attrs.items():
            if hasattr(self, kk):
                raise AttributeError(f"'Event' object already has an attribute '{kk}'")
            setattr(self, kk, vv)

    @property
    def action_information(self) -> Dataset:
        """Return an N-ACTION request's `Action Information` as a *pydicom*
        :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Action Information* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-ACTION request.
        """
        msg = (
            "The corresponding event is not an N-ACTION request and has no "
            "'Action Information' parameter"
        )
        return self._get_dataset("ActionInformation", msg)

    @property
    def action_type(self) -> int | None:
        """Return an N-ACTION request's `Action Type ID` as an :class:`int`.

        .. versionadded:: 1.4

        Returns
        -------
        int or None
            The request's (0000,1008) *Action Type ID* value, may be ``None``
            if the peer's N-ACTION request is non-conformant.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-ACTION request.
        """
        try:
            return cast("N_ACTION", self.request).ActionTypeID
        except AttributeError:
            raise AttributeError(
                "The corresponding event is not an N-ACTION request and has "
                "no 'Action Type ID' parameter"
            )

    @property
    def attribute_identifiers(self) -> list[BaseTag]:
        """Return an N-GET request's `Attribute Identifier List` as a
        :class:`list` of *pydicom* :class:`~pydicom.tag.BaseTag`.

        Returns
        -------
        list of pydicom.tag.BaseTag
            The (0000,1005) *Attribute Identifier List* tags, may be an empty
            list if no *Attribute Identifier List* was included in the C-GET
            request.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-GET request.
        """
        try:
            attr_list = cast("N_GET", self.request).AttributeIdentifierList
            if attr_list is None:
                return []

            if not isinstance(attr_list, list):
                return [attr_list]

            return attr_list
        except AttributeError:
            pass

        raise AttributeError(
            "The corresponding event is not an N-GET request and has no "
            "'Attribute Identifier List' parameter"
        )

    @property
    def attribute_list(self) -> Dataset:
        """Return an N-CREATE request's `Attribute List` as a *pydicom*
        :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Attribute List* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-CREATE request.
        """
        msg = (
            "The corresponding event is not an N-CREATE request and has no "
            "'Attribute List' parameter"
        )
        return self._get_dataset("AttributeList", msg)

    @property
    def dataset(self) -> Dataset:
        """Return a C-STORE request's `Data Set` as a *pydicom*
        :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Data Set* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not a C-STORE request.
        """
        msg = (
            "The corresponding event is not a C-STORE request and has no "
            "'Data Set' parameter"
        )
        try:
            return dcmread(self.dataset_path)
        except (TypeError, AttributeError):
            pass

        return self._get_dataset("DataSet", msg)

    @property
    def dataset_path(self) -> Path:
        """Return the path to the dataset when
        :attr:`~pynetdicom._config.STORE_RECV_CHUNKED_DATASET` is ``True``.

        .. versionadded:: 2.0

        Returns
        -------
        pathlib.Path
            The path to the dataset.
        """
        try:
            req = cast("C_STORE", self.request)
            path = req._dataset_path
        except AttributeError:
            msg = (
                "The corresponding event is either not a C-STORE request or "
                "'STORE_RECV_CHUNKED_DATASET' is not True."
            )
            raise AttributeError(msg)

        return cast(Path, path)

    def encoded_dataset(self, include_meta: bool = True) -> bytes:
        """Return the encoded C-STORE dataset sent by the peer without first
        decoding it.

        .. versionadded:: 2.1

        Examples
        --------
        Retrieve the encoded dataset as sent by the peer::

          def handle_store(event: pynetdicom.events.Event) -> int:
              stream: bytes = event.encoded_dataset(include_meta=False)

              return 0x0000

        Write the encoded dataset to file in the DICOM File Format without
        having to first decode it::

          def handle_store(event: pynetdicom.events.Event, dst: pathlib.Path) -> int:
              with dst.open("wb") as f:
                  f.write(event.encoded_dataset())

              return 0x0000

        Parameters
        ----------
        include_meta : bool, optional
            If ``True`` (default) then include the encoded DICOM preamble,
            prefix and file meta information with the returned bytestream.

        Returns
        -------
        bytes
            The encoded dataset as sent by the peer, with or without the file
            meta information.

        Raises
        ------
        AttributeError
            If the corresponding event is not a C-STORE request.
        """
        try:
            request = cast(C_STORE, self.request)
            stream = cast(BytesIO, request.DataSet).getvalue()
        except AttributeError:
            raise AttributeError(
                "The corresponding event is not a C-STORE request and has no "
                "'Data Set' parameter"
            )

        if not include_meta:
            return stream

        return b"".join((
            b"\x00" * 128,
            b"DICM",
            encode_file_meta(self.file_meta),
            stream,
        ))

    @property
    def event(self) -> EventType:
        """Return the corresponding event.

        .. versionadded:: 1.4

        Returns
        -------
        events.InterventionEvent or events.NotificationEvent
            The corresponding event as a
            :func:`namedtuple<collections.namedtuple>`.
        """
        return self._event

    @property
    def event_information(self) -> Dataset:
        """Return an N-EVENT-REPORT request's `Event Information` as a
        *pydicom* :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Event Information* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-EVENT-REPORT request.
        """
        msg = (
            "The corresponding event is not an N-EVENT-REPORT request and has "
            "no 'Event Information' parameter"
        )
        return self._get_dataset("EventInformation", msg)

    @property
    def event_type(self) -> int | None:
        """Return an N-EVENT-REPORT request's `Event Type ID` as an
        :class:`int`.

        .. versionadded:: 1.4

        Returns
        -------
        int or None
            The request's (0000,1002) *Event Type ID* value, may be ``None``
            if the peer's N-EVENT-REPORT request is non-conformant.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-EVENT-REPORT request.
        """
        try:
            return cast("N_EVENT_REPORT", self.request).EventTypeID
        except AttributeError:
            raise AttributeError(
                "The corresponding event is not an N-EVENT-REPORT request "
                "and has no 'Event Type ID' parameter"
            )

    @property
    def file_meta(self) -> FileMetaDataset:
        r"""Return a *pydicom* :class:`~pydicom.dataset.Dataset` with the
        :dcm:`File Meta Information<part10/chapter_7.html#sect_7.1>` for a
        C-STORE request's `Data Set`.

        Contains the following File Meta Information elements:

        * (0002,0000) *File Meta Information Group Length* - set as ``0``, will
          be updated with the correct value during write
        * (0002,0001) *File Meta Information Version* - set as ``0x0001``
        * (0002,0002) *Media Storage SOP Class UID* - set from the request's
          *Affected SOP Class UID*
        * (0002,0003) *Media Storage SOP Instance UID* - set from the request's
          *Affected SOP Instance UID*
        * (0002,0010) *Transfer Syntax UID* - set from the presentation context
          used to transfer the *Data Set*
        * (0002,0012) *Implementation Class UID* - set using
          :attr:`~pynetdicom.PYNETDICOM_IMPLEMENTATION_UID`
        * (0002,0013) *Implementation Version Name* - set using
          :attr:`~pynetdicom.PYNETDICOM_IMPLEMENTATION_VERSION`

        Examples
        --------

        Add the File Meta Information to the decoded *Data Set* and save it to
        the :dcm:`DICOM File Format<part10/chapter_7.html>`.

        .. code-block:: python

            >>> ds = event.dataset
            >>> ds.file_meta = event.file_meta
            >>> ds.save_as('example.dcm', write_like_original=False)

        Encode the File Meta Information in a new file and append the encoded
        *Data Set* to it. This skips having to decode/re-encode the *Data Set*
        as in the previous example (or alternatively, just use the
        :meth:`~pynetdicom.events.Event.encoded_dataset` method).

        .. code-block:: python

           >>> from pydicom.filewriter import write_file_meta_info
           >>> with open('example.dcm', 'wb') as f:
           ...     f.write(b'\x00' * 128)
           ...     f.write(b'DICM')
           ...     write_file_meta_info(f, event.file_meta)
           ...     f.write(event.request.DataSet.getvalue())

        Returns
        -------
        pydicom.dataset.Dataset
            The File Meta Information suitable for use with the decoded C-STORE
            request's *Data Set*.

        Raises
        ------
        AttributeError
            If the corresponding event is not a C-STORE request.
        """
        if not hasattr(self.request, "DataSet"):
            raise AttributeError("The corresponding event is not a C-STORE request")

        # A C-STORE request must have AffectedSOPClassUID and
        #   AffectedSOPInstanceUID
        self.request = cast("C_STORE", self.request)
        sop_class = cast(UID, self.request.AffectedSOPClassUID)
        sop_instance = cast(UID, self.request.AffectedSOPInstanceUID)
        return create_file_meta(
            sop_class_uid=sop_class,
            sop_instance_uid=sop_instance,
            transfer_syntax=self.context.transfer_syntax,
        )

    def _get_dataset(self, attr: str, exc_msg: str) -> Dataset:
        """Return DIMSE dataset-like parameter as a *pydicom* Dataset.

        Parameters
        ----------
        attr : str
            The name of the DIMSE primitive's dataset-like parameter, one of
            'DataSet', 'Identifier', 'AttributeList', 'ModificationList',
            'EventInformation', 'ActionInformation'.
        exc_msg : str
            The exception message to use if the request primitive has no
            dataset-like parameter.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded dataset-like parameter.

        Raises
        ------
        AttributeError
            If the corresponding event is not due to one of the DIMSE requests
            with a dataset-like parameter.
        """
        try:
            bytestream = getattr(self.request, attr)

            # If no change in encoded data then return stored decode
            if self._hash == hash(bytestream):
                return cast(Dataset, self._decoded)

            # Some dataset-like parameters are optional
            if bytestream and bytestream.getvalue() != b"":
                # Dataset-like parameter has been used
                t_syntax = self.context.transfer_syntax
                ds = decode(
                    bytestream,
                    t_syntax.is_implicit_VR,
                    t_syntax.is_little_endian,
                    t_syntax.is_deflated,
                )

                ds.is_little_endian = t_syntax.is_little_endian
                ds.is_implicit_VR = t_syntax.is_implicit_VR

                # Store the decoded dataset in case its accessed again
                self._decoded = ds
            else:
                # Dataset-like parameter hasn't been used
                self._decoded = Dataset()

            self._hash = hash(bytestream)
            return self._decoded

        except AttributeError:
            pass

        raise AttributeError(exc_msg)

    @property
    def identifier(self) -> Dataset:
        """Return a C-FIND, C-GET or C-MOVE request's `Identifier` as a
        *pydicom* :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Identifier* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not a C-FIND, C-GET or C-MOVE
            request.
        """
        msg = (
            "The corresponding event is not a C-FIND, C-GET or C-MOVE request "
            "and has no 'Identifier' parameter"
        )
        return self._get_dataset("Identifier", msg)

    @property
    def is_cancelled(self) -> bool:
        """Return ``True`` if a C-CANCEL request has been received.

        Returns
        -------
        bool
            If this event corresponds to a C-FIND, C-GET or C-MOVE request
            being received by a Service Class then returns ``True`` if a
            C-CANCEL request with a *Message ID Being Responded To* parameter
            value corresponding to the *Message ID* of the service request has
            been received. If no such C-CANCEL request has been received or if
            the event is not a C-FIND, C-GET or C-MOVE request then returns
            ``False``.
        """
        try:
            return self._is_cancelled(cast(int, self.request.MessageID))
        except AttributeError:
            pass

        return False

    @property
    def message_id(self) -> int:
        """Return a DIMSE service request's `Message ID` as :class:`int`.

        .. versionadded:: 1.5

        Returns
        -------
        int
            The request's (0000,0110) *Message ID* value.

        Raises
        ------
        AttributeError
            If the corresponding event is not one of the DIMSE service
            requests.
        """
        try:
            return cast(int, self.request.MessageID)
        except AttributeError:
            raise AttributeError(
                "The corresponding event is not a DIMSE service request and "
                "has no 'Message ID' parameter"
            )

    @property
    def modification_list(self) -> Dataset:
        """Return an N-SET request's `Modification List` as a *pydicom*
        :class:`~pydicom.dataset.Dataset`.

        Because *pydicom* defers data parsing during decoding until an element
        is actually required the returned :class:`~pydicom.dataset.Dataset`
        may raise an exception when any element is first accessed. It's
        therefore important that proper error handling be part of any handler
        that uses the returned :class:`~pydicom.dataset.Dataset`.

        Returns
        -------
        pydicom.dataset.Dataset
            The decoded *Modification List* dataset.

        Raises
        ------
        AttributeError
            If the corresponding event is not an N-SET request.
        """
        msg = (
            "The corresponding event is not an N-SET request and has no "
            "'Modification List' parameter"
        )
        return self._get_dataset("ModificationList", msg)

    @property
    def move_destination(self) -> str | None:
        """Return a C-MOVE request's `Move Destination` as :class:`str`.

        .. versionadded:: 1.4

        .. versionchanged:: 2.0

            Changed to return :class:`str` and trailing spaces removed.

        Returns
        -------
        str or None
            The request's (0000,0600) *Move Destination* value, without any
            trailing padding spaces. May be ``None`` if the peer's C-MOVE
            request is non-conformant.

        Raises
        ------
        AttributeError
            If the corresponding event is not a C-MOVE request.
        """
        try:
            return cast("C_MOVE", self.request).MoveDestination
        except AttributeError:
            raise AttributeError(
                "The corresponding event is not a C-MOVE request and has no "
                "'Move Destination' parameter"
            )


# Default extended negotiation event handlers
def _async_ops_handler(event: Event) -> tuple[int, int]:
    """Default handler for when an Asynchronous Operations Window Negotiation
    item is include in the association request.

    See _handlers.doc_handle_async for detailed documentation.
    """
    raise NotImplementedError(
        "No handler has been bound to 'evt.EVT_ASYNC_OPS', so no "
        "Asynchronous Operations Window Negotiation response will be "
        "sent"
    )


def _sop_common_handler(event: Event) -> dict[UID, "SOPClassCommonExtendedNegotiation"]:
    """Default handler for when one or more SOP Class Common Extended
    Negotiation items are included in the association request.

    See _handlers.doc_handle_sop_common for detailed documentation.
    """
    return {}


def _sop_extended_handler(event: Event) -> dict[UID, bytes]:
    """Default handler for when one or more SOP Class Extended Negotiation
    items are included in the association request.

    See _handlers.doc_handler_sop_extended for detailed documentation.
    """
    return {}


def _user_identity_handler(event: Event) -> tuple[bool, bytes | None]:
    """Default handler for when a user identity negotiation item is included
    with the association request.

    See _handlers.doc_handler_userid for detailed documentation.
    """
    raise NotImplementedError(
        "No handler has been bound to 'evt.EVT_USER_ID', so the User Identity "
        "Negotiation will be ignored and the association accepted (unless "
        "rejected for another reason)"
    )


# Default service class request handlers
def _c_echo_handler(event: Event) -> _BasicReturnType:
    """Default handler for when a C-ECHO request is received.

    See _handlers.doc_handle_echo for detailed documentation.
    """
    return 0x0000


def _c_find_handler(event: Event) -> _IteratorType:
    """Default handler for when a C-FIND request is received.

    See _handlers.doc_handle_find for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_C_FIND'")


def _c_get_handler(event: Event) -> _IteratorType:
    """Default handler for when a C-GET request is received.

    See _handlers.doc_handle_c_get for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_C_GET'")


def _c_move_handler(event: Event) -> _IteratorType:
    """Default handler for when a C-MOVE request is received.

    See _handlers.doc_handle_move for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_C_MOVE'")


def _c_store_handler(event: Event) -> _BasicReturnType:
    """Default handler for when a C-STORE request is received.

    See _handlers.doc_handle_store for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_C_STORE'")


def _n_action_handler(event: Event) -> _DatasetReturnType:
    """Default handler for when an N-ACTION request is received.

    See _handlers.doc_handle_action for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_ACTION'")


def _n_create_handler(event: Event) -> _DatasetReturnType:
    """Default handler for when an N-CREATE request is received.

    See _handlers.doc_handle_create for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_CREATE'")


def _n_delete_handler(event: Event) -> _BasicReturnType:
    """Default handler for when an N-DELETE request is received.

    See _handlers.doc_handle_delete for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_DELETE'")


def _n_event_report_handler(event: Event) -> _DatasetReturnType:
    """Default handler for when an N-EVENT-REPORT request is received.

    See _handlers.doc_handle_event_report for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_EVENT_REPORT'")


def _n_get_handler(event: Event) -> _DatasetReturnType:
    """Default handler for when an N-GET request is received.

    See _handlers.doc_handle_n_get for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_GET'")


def _n_set_handler(event: Event) -> _DatasetReturnType:
    """Default handler for when an N-SET request is received.

    See _handlers.doc_handle_set for detailed documentation.
    """
    raise NotImplementedError("No handler has been bound to 'evt.EVT_N_SET'")



"""Unit tests for the QRSCP app's database functions."""

import os
from pathlib import Path
import sys
import tempfile

import pytest

try:
    from sqlalchemy import create_engine
    from sqlalchemy.schema import MetaData
    from sqlalchemy.exc import IntegrityError, SAWarning
    from sqlalchemy.orm import sessionmaker

    HAVE_SQLALCHEMY = True
except ImportError:
    HAVE_SQLALCHEMY = False

from pydicom import dcmread
import pydicom.config
from pydicom.dataset import Dataset
from pydicom.tag import Tag

from pynetdicom.sop_class import (
    PatientRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelFind,
)

if HAVE_SQLALCHEMY:
    from pynetdicom.apps.qrscp import db


TEST_DIR = Path(__file__).parent
DATA_DIR = TEST_DIR.parent.parent / "tests" / "dicom_files"
DATASETS = {
    "CTImageStorage.dcm": {
        "patient_id": "1CT1",
        "patient_name": "CompressedSamples^CT1",
        "study_instance_uid": "1.3.6.1.4.1.5962.1.2.1.20040119072730.12322",
        "study_date": "20040119",
        "study_time": "072730",
        "accession_number": None,
        "study_id": "1CT1",
        "series_instance_uid": "1.3.6.1.4.1.5962.1.3.1.1.20040119072730.12322",
        "modality": "CT",
        "series_number": "1",
        "sop_instance_uid": "1.3.6.1.4.1.5962.1.1.1.1.1.20040119072730.12322",
        "instance_number": "1",
        "transfer_syntax_uid": "1.2.840.10008.1.2.1",
        "sop_class_uid": "1.2.840.10008.5.1.4.1.1.2",
    },
    "MRImageStorage_ExplicitVRBigEndian.dcm": {
        "patient_id": "4MR1",
        "patient_name": "CompressedSamples^MR1",
        "study_instance_uid": "1.3.6.1.4.1.5962.1.2.4.20040826185059.5457",
        "study_date": "20040826",
        "study_time": "185059",
        "accession_number": None,
        "study_id": "4MR1",
        "series_instance_uid": "1.3.6.1.4.1.5962.1.3.4.1.20040826185059.5457",
        "modality": "MR",
        "series_number": "1",
        "sop_instance_uid": "1.3.6.1.4.1.5962.1.1.4.1.1.20040826185059.5457.1",
        "instance_number": "1",
        "transfer_syntax_uid": "1.2.840.10008.1.2.4.90",
        "sop_class_uid": "1.2.840.10008.5.1.4.1.1.4",
    },
    "MRImageStorage_JPG2000_Lossless.dcm": {
        "patient_id": "4MR1",
        "patient_name": "CompressedSamples^MR1",
        "study_instance_uid": "1.3.6.1.4.1.5962.1.2.4.20040826185059.5457",
        "study_date": "20040826",
        "study_time": "185059",
        "accession_number": None,
        "study_id": "4MR1",
        "series_instance_uid": "1.3.6.1.4.1.5962.1.3.4.1.20040826185059.5457",
        "modality": "MR",
        "series_number": "1",
        "sop_instance_uid": "1.3.6.1.4.1.5962.1.1.4.1.1.20040826185059.5457",
        "instance_number": "1",
        "transfer_syntax_uid": "1.2.840.10008.1.2.4.90",
        "sop_class_uid": "1.2.840.10008.5.1.4.1.1.4",
    },
    "RTImageStorage.dcm": {
        "patient_id": "0123456789",
        "patient_name": "ANON^A^B^C^D",
        "study_instance_uid": "1.3.46.423632.132218.1415242681.6",
        "study_date": "20150803",
        "study_time": "114426",
        "accession_number": None,
        "study_id": "00000000",
        "series_instance_uid": "1.3.46.423632.132218.1415243125.23",
        "modality": "RTIMAGE",
        "series_number": "000000052637",
        "sop_instance_uid": "1.3.46.423632.132218.1438566266.11",
        "instance_number": "000000056675",
        "transfer_syntax_uid": "1.2.840.10008.1.2",
        "sop_class_uid": "1.2.840.10008.5.1.4.1.1.481.1",
    },
    "SCImageStorage_Deflated.dcm": {
        "patient_id": None,
        "patient_name": "^^^^",
        "study_instance_uid": "1.3.6.1.4.1.5962.1.2.0.977067310.6001.0",
        "study_date": None,
        "study_time": None,
        "accession_number": None,
        "study_id": None,
        "series_instance_uid": "1.3.6.1.4.1.5962.1.3.0.0.977067310.6001.0",
        "modality": "OT",
        "series_number": None,
        "sop_instance_uid": "1.3.6.1.4.1.5962.1.1.0.0.0.977067309.6001.0",
        "instance_number": None,
        "transfer_syntax_uid": "1.2.840.10008.1.2.1.99",
        "sop_class_uid": "1.2.840.10008.5.1.4.1.1.7",
    },
}


try:
    pydicom.config.settings.writing_validation_mode = 0
    pydicom.config.settings.reading_validation_mode = 0
except AttributeError:
    pass


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestConnect:
    """Tests for db.connect()."""

    def test_create_new(self):
        """Test connecting to the instance database if it doesn't exist."""
        db_file = tempfile.NamedTemporaryFile()
        db_location = f"sqlite:///{db_file.name}"
        engine = db.create(db_location)

        # Check exists with tables
        meta = MetaData()
        meta.reflect(bind=engine)
        assert "patient" in meta.tables
        assert "study" in meta.tables
        assert "series" in meta.tables
        assert "image" in meta.tables
        assert "instance" in meta.tables

    def test_create_new_existing(self):
        """Test connecting to the instance database if it exists."""
        db_file = tempfile.NamedTemporaryFile()
        db_location = "sqlite:///{}".format(db_file.name)
        db.create(db_location)
        # Test creating if already exists
        engine = db.create(db_location)

        meta = MetaData()
        meta.reflect(bind=engine)
        assert "patient" in meta.tables
        assert "study" in meta.tables
        assert "series" in meta.tables
        assert "image" in meta.tables
        assert "instance" in meta.tables


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestAddInstance:
    """Tests for db.add_instance()."""

    def setup_method(self):
        """Run prior to each test"""
        engine = db.create("sqlite:///:memory:")

        pydicom.config.use_none_as_empty_text_VR_value = True

        ds = Dataset()
        ds.PatientID = "1234"
        ds.StudyInstanceUID = "1.2"
        ds.SeriesInstanceUID = "1.2.3"
        ds.SOPInstanceUID = "1.2.3.4"
        self.minimal = ds

        self.session = sessionmaker(bind=engine)()

    def test_add_instance(self):
        """Test adding to the instance database."""
        ds = dcmread(os.fspath(DATA_DIR / "CTImageStorage.dcm"))
        db.add_instance(ds, self.session)

        obj = self.session.query(db.Instance).all()
        assert 1 == len(obj)
        for kk, vv in DATASETS["CTImageStorage.dcm"].items():
            assert vv == getattr(obj[0], kk)

    def test_add_multiple_instances(self):
        """Test adding multiple data to the instance database."""
        for fname in DATASETS:
            ds = dcmread(os.fspath(DATA_DIR / fname))
            db.add_instance(ds, self.session)

        obj = self.session.query(db.Instance).all()
        assert 5 == len(obj)
        obj = self.session.query(db.Instance, db.Instance.patient_name).all()
        names = [val[1] for val in obj]
        assert "CompressedSamples^CT1" in names
        assert "CompressedSamples^MR1" in names
        assert "ANON^A^B^C^D" in names
        assert "^^^^" in names

    def test_add_minimal(self):
        """Test adding a minimal dataset."""
        db.add_instance(self.minimal, self.session)

        obj = self.session.query(db.Instance).all()
        assert 1 == len(obj)
        assert "1234" == obj[0].patient_id
        assert "1.2" == obj[0].study_instance_uid
        assert "1.2.3" == obj[0].series_instance_uid
        assert "1.2.3.4" == obj[0].sop_instance_uid

        rest = [
            "patient_name",
            "study_date",
            "study_time",
            "accession_number",
            "study_id",
            "modality",
            "series_number",
            "instance_number",
        ]
        for attr in rest:
            assert getattr(obj[0], attr) is None

    def test_bad_instance(self):
        """Test that instances with bad data aren't added."""
        keywords = [
            ("PatientID", 64),
            ("PatientName", 400),
            ("StudyInstanceUID", 64),
            ("StudyDate", 8),
            ("StudyTime", 14),
            ("AccessionNumber", 16),
            ("StudyID", 16),
            ("SeriesInstanceUID", 64),
            ("Modality", 16),
            # ('SeriesNumber', None),
            ("SOPInstanceUID", 64),
            # ('InstanceNumber', None),
        ]
        ds = self.minimal[:]
        for kw, max_len in keywords:
            setattr(ds, kw, "a" * (max_len + 1))
            with pytest.raises(AssertionError):
                db.add_instance(ds, self.session)

        assert not self.session.query(db.Instance).all()

    def test_bad_instance_none(self):
        """Test that instances with bad data aren't added."""
        keywords = [
            ("PatientID", 64),
            ("PatientName", 64),
            ("StudyInstanceUID", 64),
            ("StudyDate", 8),
            ("StudyTime", 14),
            ("AccessionNumber", 16),
            ("StudyID", 16),
            ("SeriesInstanceUID", 64),
            ("Modality", 16),
            # ('SeriesNumber', None),
            ("SOPInstanceUID", 64),
            # ('InstanceNumber', None),
        ]
        ds = Dataset()
        ds.PatientID = None
        ds.StudyInstanceUID = None
        ds.SeriesInstanceUID = None
        ds.SOPInstanceUID = None
        with pytest.raises(IntegrityError):
            msg = r"Column 'instance.sop_instance_uid' is marked as a"
            with pytest.warns(SAWarning, match=msg):
                db.add_instance(ds, self.session)

        self.session.rollback()

        assert not self.session.query(db.Instance).all()

    def test_instance_exists(self):
        """Test that adding already existing instance updates it."""
        db.add_instance(self.minimal, self.session)

        result = self.session.query(db.Instance).all()
        assert 1 == len(result)
        assert None == result[0].modality

        self.minimal.Modality = "CT"

        db.add_instance(self.minimal, self.session)
        result = self.session.query(db.Instance).all()
        assert 1 == len(result)
        assert "CT" == result[0].modality


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestRemoveInstance:
    """Tests for db.remove_instance()."""

    def setup_method(self):
        """Run prior to each test"""
        engine = db.create("sqlite:///:memory:")

        pydicom.config.use_none_as_empty_text_VR_value = True

        self.session = sessionmaker(bind=engine)()
        ds = Dataset()
        ds.PatientID = "1234"
        ds.StudyInstanceUID = "1.2"
        ds.SeriesInstanceUID = "1.2.3"
        ds.SOPInstanceUID = "1.2.3.4"
        db.add_instance(ds, self.session)

        self.minimal = ds

    def test_remove_existing(self):
        """Test removing if exists in database."""

        obj = self.session.query(db.Instance).all()
        assert 1 == len(obj)
        assert "1.2.3.4" == obj[0].sop_instance_uid

        db.remove_instance("1.2.3.4", self.session)

        assert not self.session.query(db.Instance).all()
        assert not self.session.query(db.Patient).all()
        assert not self.session.query(db.Study).all()
        assert not self.session.query(db.Series).all()
        assert not self.session.query(db.Image).all()

    def test_remove_not_existing(self):
        """Test removing if doesn't exist in database."""

        obj = self.session.query(db.Instance).all()
        assert 1 == len(obj)
        assert "1.2.3.4" == obj[0].sop_instance_uid

        db.remove_instance("1.2.3.5", self.session)

        assert self.session.query(db.Instance).all()


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestClear:
    """Tests for db.clear()."""

    def setup_method(self):
        """Run prior to each test"""
        engine = db.create("sqlite:///:memory:")

        pydicom.config.use_none_as_empty_text_VR_value = True

        self.session = sessionmaker(bind=engine)()
        for fname in DATASETS:
            ds = dcmread(os.fspath(DATA_DIR / fname))
            db.add_instance(ds, self.session)

    def test_clear(self):
        """Test removing if exists in database."""
        assert 5 == len(self.session.query(db.Instance).all())

        db.clear(self.session)
        assert not self.session.query(db.Instance).all()


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestSearch:
    """Tests for db.search()."""

    def setup_method(self):
        """Run prior to each test"""
        engine = db.create("sqlite:///:memory:")
        pydicom.config.use_none_as_empty_text_VR_value = True

        self.session = sessionmaker(bind=engine)()

        for fname in DATASETS:
            ds = dcmread(os.fspath(DATA_DIR / fname))
            db.add_instance(ds, self.session)

    def test_search(self):
        """Test simple search."""
        pass

    def test_search_range_both(self):
        """Test searching a range with both start and end."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.StudyDate = "20000101-20200101"

        q = db._search_range(query["StudyDate"], self.session)
        assert 4 == len(q.all())

        query.StudyDate = "20000101-20150101"
        q = db._search_range(query["StudyDate"], self.session)
        assert 3 == len(q.all())

        query.StudyDate = "20000101-20010101"
        q = db._search_range(query["StudyDate"], self.session)
        assert not q.all()

    def test_search_range_both_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_range_start(self):
        """Test searching a range with only start."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.StudyDate = "20000101-"

        q = db._search_range(query["StudyDate"], self.session)
        assert 4 == len(q.all())

        query.StudyDate = "20150101-"
        q = db._search_range(query["StudyDate"], self.session)
        assert 1 == len(q.all())

        query.StudyDate = "20200101-"
        q = db._search_range(query["StudyDate"], self.session)
        assert not q.all()

    def test_search_range_start_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_range_end(self):
        """Test searching a range with only end."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.StudyDate = "-20200101"

        q = db._search_range(query["StudyDate"], self.session)
        assert 4 == len(q.all())

        query.StudyDate = "-20150101"
        q = db._search_range(query["StudyDate"], self.session)
        assert 3 == len(q.all())

        query.StudyDate = "-20010101"
        q = db._search_range(query["StudyDate"], self.session)
        assert not q.all()

    def test_search_range_end_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_range_neither(self):
        """Test searching a range with only end."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.StudyDate = "-"

        with pytest.raises(ValueError):
            db._search_range(query["StudyDate"], self.session)

    def test_search_range_neither_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_single_value(self):
        """Test search using a single value."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientName = "CompressedSamples^CT1"

        q = db._search_single_value(query["PatientName"], self.session)
        assert 1 == len(q.all())

    def test_search_single_value_subquery(self):
        """Test searching using a single value within an existing query."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientName = None

        q = db._search_universal(query["PatientName"], self.session)
        assert 5 == len(q.all())

        query.PatientName = "CompressedSamples^CT1"
        q = db._search_single_value(query["PatientName"], self.session, q)
        assert 1 == len(q.all())

    def test_search_wildcard_asterisk(self):
        """Test search using a * wildcard."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientName = "*"

        q = db._search_wildcard(query["PatientName"], self.session)
        assert 5 == len(q.all())

        query.PatientName = "CompressedSamples*"
        q = db._search_wildcard(query["PatientName"], self.session)
        assert 3 == len(q.all())

    def test_search_wildcard_asterisk_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_wildcard_qmark(self):
        """Test search using a ? wildcard."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientName = "CompressedSamples^??1"

        q = db._search_wildcard(query["PatientName"], self.session)
        assert 3 == len(q.all())

    def test_search_wildcard_qmark_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_uid_list(self):
        """Test search using a UID list."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.SOPInstanceUID = None

        q = db._search_uid_list(query["SOPInstanceUID"], self.session)
        assert 5 == len(q.all())

        query.SOPInstanceUID = ["1.3.6.1.4.1.5962.1.1.4.1.1.20040826185059.5457"]
        q = db._search_uid_list(query["SOPInstanceUID"], self.session)
        assert 1 == len(q.all())

        query.SOPInstanceUID = [
            "1.3.6.1.4.1.5962.1.1.4.1.1.20040826185059.5457",
            "1.3.6.1.4.1.5962.1.1.1.1.1.20040119072730.12322",
        ]
        q = db._search_uid_list(query["SOPInstanceUID"], self.session)
        assert 2 == len(q.all())

        query.SOPInstanceUID = [
            "1.3.6.1.4.1.5962.1.1.4.1.1.20040826185059.5457",
            "1.3.6.1.4.1.5962.1.1.1.1.1.20040119072730.12322",
            "1.3.46.423632.132218.1415242681.6",
        ]
        q = db._search_uid_list(query["SOPInstanceUID"], self.session)
        assert 2 == len(q.all())

    def test_search_uid_list_subquery(self):
        """Test searching within an existing query."""
        pass

    def test_search_uid_list_empty(self):
        """Test searching an empty UID element works correctly."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.SOPInstanceUID = None

        q = db._search_uid_list(query["SOPInstanceUID"], self.session)
        assert 5 == len(q.all())

    @pytest.mark.skip("probably obsolete")
    def test_combine_queries(self):
        """Test combining queries."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientName = "CompressedSamples*"
        query.StudyDate = "20000101-20040119"

        q1 = db._search_wildcard(query["PatientName"], self.session)
        q2 = db._search_range(query["StudyDate"], self.session)
        q = q1.intersect(q2)
        assert 1 == len(q.all())


IDENTIFIERS = [
    (
        "PATIENT",
        [],
        r"The Identifier contains no keys",
        r"The Identifier's Query Retrieve Level value is invalid",
    ),
    (
        "PATIENT",
        ["PatientName"],
        None,
        r"The Identifier's Query Retrieve Level value is invalid",
    ),
    (
        "PATIENT",
        ["PatientID"],
        None,
        r"The Identifier's Query Retrieve Level value is invalid",
    ),
    (
        "PATIENT",
        ["PatientID", "SOPInstanceUID"],
        r"contains keys below the level specified by the Query Retrieve",
        r"The Identifier's Query Retrieve Level value is invalid",
    ),
    (
        "STUDY",
        [],
        r"The Identifier contains no keys",
        r"The Identifier contains no keys",
    ),
    ("STUDY", ["PatientName"], r"missing a unique key for the 'PATIENT' level", None),
    ("STUDY", ["PatientID"], None, None),
    ("STUDY", ["PatientID", "PatientName"], None, None),
    ("STUDY", ["PatientID", "StudyInstanceUID"], None, None),
    (
        "STUDY",
        ["PatientID", "StudyInstanceUID", "SOPInstanceUID"],
        r"contains keys below the level specified by the Query Retrieve",
        r"contains keys below the level specified by the Query Retrieve",
    ),
    (
        "SERIES",
        [],
        r"The Identifier contains no keys",
        r"The Identifier contains no keys",
    ),
    (
        "SERIES",
        ["PatientID"],
        r"missing a unique key for the 'STUDY' level",
        r"missing a unique key for the 'STUDY' level",
    ),
    (
        "SERIES",
        ["StudyInstanceUID"],
        r"missing a unique key for the 'PATIENT' level",
        None,
    ),
    ("SERIES", ["PatientID", "StudyInstanceUID"], None, None),
    (
        "SERIES",
        ["PatientID", "StudyInstanceUID", "SOPInstanceUID"],
        r"contains keys below the level specified by the Query Retrieve",
        r"contains keys below the level specified by the Query Retrieve",
    ),
    (
        "IMAGE",
        [],
        r"The Identifier contains no keys",
        r"The Identifier contains no keys",
    ),
    (
        "IMAGE",
        ["PatientID"],
        r"missing a unique key for the 'STUDY' level",
        r"missing a unique key for the 'STUDY' level",
    ),
    (
        "IMAGE",
        ["PatientID", "StudyInstanceUID"],
        r"missing a unique key for the 'SERIES' level",
        r"missing a unique key for the 'SERIES' level",
    ),
    ("IMAGE", ["PatientID", "StudyInstanceUID", "SeriesInstanceUID"], None, None),
    (
        "DUMMY",
        [],
        r"The Identifier's Query Retrieve Level value is invalid",
        r"The Identifier's Query Retrieve Level value is invalid",
    ),
]


@pytest.mark.skipif(not HAVE_SQLALCHEMY, reason="Requires sqlalchemy")
class TestSearchFind:
    """Tests for running C-FIND queries against the database."""

    def setup_method(self):
        """Run prior to each test"""
        engine = db.create("sqlite:///:memory:")
        pydicom.config.use_none_as_empty_text_VR_value = True

        self.session = sessionmaker(bind=engine)()
        for fname in DATASETS:
            ds = dcmread(DATA_DIR / fname)
            db.add_instance(ds, self.session)

    def test_no_attributes(self):
        """Test searching with no attributes."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"

        model = PatientRootQueryRetrieveInformationModelFind

        msg = r"The Identifier contains no keys"
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._search_qr(model, query, self.session)

    def test_patient_minimal(self):
        """Test query at PATIENT level for Patient Root."""
        query = Dataset()
        query.QueryRetrieveLevel = "PATIENT"
        query.PatientID = None
        query.PatientName = "CompressedSamples^??1"

        model = PatientRootQueryRetrieveInformationModelFind

        result = db._search_qr(model, query, self.session)
        assert 3 == len(result)

    def test_check_identifier_patient(self):
        """Tests for check_find_identifier()."""
        ds = Dataset()
        ds.QueryRetrieveLevel = "PATIENT"
        ds.PatientName = None
        patient = PatientRootQueryRetrieveInformationModelFind

        db._check_identifier(ds, patient)

        ds.QueryRetrieveLevel = "STUDY"
        msg = r"The Identifier is missing a unique key for the 'PATIENT' level"
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._check_identifier(ds, patient)

        ds.PatientID = "12345"
        db._check_identifier(ds, patient)

    @pytest.mark.parametrize("level, identifier, pt, st", IDENTIFIERS)
    def test_check_identifier_param_patient(self, level, identifier, pt, st):
        """Tests for check_identifier()."""
        ds = Dataset()
        ds.QueryRetrieveLevel = level
        for kw in identifier:
            setattr(ds, kw, None)

        model = PatientRootQueryRetrieveInformationModelFind
        if pt:
            with pytest.raises(db.InvalidIdentifier, match=pt):
                db._check_identifier(ds, model)
        else:
            db._check_identifier(ds, model)

    def test_check_identifier_study(self):
        """Tests for check_find_identifier()."""
        ds = Dataset()
        ds.QueryRetrieveLevel = "PATIENT"
        ds.PatientName = None
        study = StudyRootQueryRetrieveInformationModelFind

        msg = r"The Identifier's Query Retrieve Level value is invalid"
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._check_identifier(ds, study)

        ds.QueryRetrieveLevel = "STUDY"
        db._check_identifier(ds, study)

        ds.QueryRetrieveLevel = "SERIES"
        msg = r"The Identifier is missing a unique key for the 'STUDY' level"
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._check_identifier(ds, study)
        ds.PatientID = None
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._check_identifier(ds, study)

        ds.StudyInstanceUID = "12345"
        db._check_identifier(ds, study)

    @pytest.mark.parametrize("level, identifier, pt, st", IDENTIFIERS)
    def test_check_identifier_param_study(self, level, identifier, pt, st):
        """Tests for check_identifier()."""
        ds = Dataset()
        ds.QueryRetrieveLevel = level
        for kw in identifier:
            setattr(ds, kw, None)

        model = StudyRootQueryRetrieveInformationModelFind
        if st:
            with pytest.raises(db.InvalidIdentifier, match=st):
                db._check_identifier(ds, model)
        else:
            db._check_identifier(ds, model)

    def test_check_identifier_raises(self):
        """Check raises if no QueryRetrieveLevel element."""
        model = PatientRootQueryRetrieveInformationModelFind
        msg = r"The Identifier contains no Query Retrieve Level element"
        with pytest.raises(db.InvalidIdentifier, match=msg):
            db._check_identifier(Dataset(), model)



"""Unit tests for the apps and pynetdicom.apps.common module."""

from collections import namedtuple
import logging
import os

import pytest

try:
    import pyfakefs

    HAVE_PYFAKEFS = True
except ImportError:
    HAVE_PYFAKEFS = False

from pydicom.dataset import Dataset
from pydicom.tag import Tag

from pynetdicom.apps.common import ElementPath, create_dataset, get_files


class TestCreateDataset:
    """Tests for pynetdicom.apps.common.create_dataset()."""

    def test_element_new(self):
        """Test creating an element using keywords."""
        Args = namedtuple("args", ["keyword", "file"])
        args = Args(["PatientName=*"], None)
        ds = create_dataset(args)

        ref = Dataset()
        ref.PatientName = "*"

        assert ds == ref

    def test_sequence_new(self):
        """Test creating a sequence using keywords."""
        Args = namedtuple("args", ["keyword", "file"])
        args = Args(["BeamSequence[0].PatientName=*"], None)
        ds = create_dataset(args)

        ref = Dataset()
        ref.BeamSequence = [Dataset()]
        ref.BeamSequence[0].PatientName = "*"

        assert ds == ref

    def test_keyword_exception(self):
        """Test exception is raised by bad keyword."""
        Args = namedtuple("args", ["keyword", "file"])
        args = Args(["BadPatientName=*"], None)

        msg = r"Unable to parse element path component"
        with pytest.raises(ValueError, match=msg):
            create_dataset(args)

    def test_file_read_exception(self):
        """Test exception is raised by bad file read."""
        Args = namedtuple("args", ["keyword", "file"])
        args = Args([], "no_such_file")

        # General exception
        msg = r"No such file or directory"
        with pytest.raises(Exception, match=msg):
            create_dataset(args)


class TestElementPath:
    """Tests for utils.ElementPath."""

    def test_child(self):
        """Tests for ElementPath.child."""
        elem = ElementPath("(0000,0000)")
        assert elem.child is None
        elem = ElementPath("CommandGroupLength")
        assert elem.child is None
        elem = ElementPath("(0000,0000).(0000,0002)")
        assert elem.child is not None
        assert elem.components == ["(0000,0000)", "(0000,0002)"]
        child = elem.child
        assert child.child is None
        assert child.components == ["(0000,0002)"]

    def test_parent(self):
        """Tests for ElementPath.parent."""
        elem = ElementPath("(0000,0000)")
        assert elem.parent is None
        elem = ElementPath("CommandGroupLength")
        assert elem.parent is None
        elem = ElementPath("(0000,0000).(0000,0002)")
        assert elem.child is not None
        assert elem.parent is None
        assert elem.components == ["(0000,0000)", "(0000,0002)"]
        child = elem.child
        assert child.parent == elem
        assert child.child is None

    def test_item_nr(self):
        """Tests for ElementPath.item_nr."""
        elem = ElementPath("PatientName")
        assert elem.is_sequence is False
        assert elem.item_nr is None

        elem = ElementPath("PatientName[0]")
        assert elem.is_sequence
        assert elem.item_nr == 0

        elem = ElementPath("PatientName[12]")
        assert elem.is_sequence
        assert elem.item_nr == 12

    def test_is_sequence(self):
        """Tests for ElementPath.is_sequence."""
        elem = ElementPath("PatientName[0]")
        assert elem.is_sequence
        assert elem.item_nr == 0

        elem = ElementPath("PatientName[12]")
        assert elem.is_sequence
        assert elem.item_nr == 12

    def test_is_sequence_raises(self):
        """Tests for ElementPath.is_sequence with invalid values."""
        msg = r"Element path contains an invalid component:"
        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName[")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName][")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName[]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName[-1]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName[-10]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("PatientName[as]")

    def test_keyword(self):
        """Tests for ElementPath.keyword."""
        elem = ElementPath("PatientName")
        assert elem.keyword == "PatientName"
        elem = ElementPath("0020,0020")
        assert elem.keyword == "PatientOrientation"
        elem = ElementPath("7ffe,0020")
        assert elem.keyword == "VariableCoefficientsSDVN"
        elem = ElementPath("0009,0020")
        assert elem.keyword == "Unknown"

    def test_non_sequence(self):
        """Test ElementPath using a non-sequence component."""
        elem = ElementPath("(0000,0000)")
        assert elem.tag == Tag(0x0000, 0x0000)
        assert elem.keyword == "CommandGroupLength"
        assert elem.VR == "UL"
        assert elem.item_nr is None
        assert elem.is_sequence is False
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

        elem = ElementPath("0000,0000")
        assert elem.tag == Tag(0x0000, 0x0000)
        assert elem.keyword == "CommandGroupLength"
        assert elem.VR == "UL"
        assert elem.item_nr is None
        assert elem.is_sequence is False
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

        elem = ElementPath("CommandGroupLength")
        assert elem.tag == Tag(0x0000, 0x0000)
        assert elem.keyword == "CommandGroupLength"
        assert elem.VR == "UL"
        assert elem.item_nr is None
        assert elem.is_sequence is False
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

    def test_pathing(self):
        """Test ElementPath using normal pathing."""
        paths = ["(0000,0000)=", "0000,0000=", "CommandGroupLength="]
        for path in paths:
            elem = ElementPath(path)
            assert elem.tag == Tag(0x0000, 0x0000)
            assert elem.keyword == "CommandGroupLength"
            assert elem.VR == "UL"
            assert elem.item_nr is None
            assert elem.is_sequence is False
            assert elem.child is None
            assert elem.parent is None
            assert elem.value == ""

        paths = ["(0000,0000)=120", "0000,0000=120", "CommandGroupLength=120"]
        for path in paths:
            elem = ElementPath(path)
            assert elem.tag == Tag(0x0000, 0x0000)
            assert elem.keyword == "CommandGroupLength"
            assert elem.VR == "UL"
            assert elem.item_nr is None
            assert elem.is_sequence is False
            assert elem.child is None
            assert elem.parent is None
            assert elem.value == 120

        paths = ["(300a,00b0)[0]=", "300a,00b0[0]=", "BeamSequence[0]="]
        for path in paths:
            elem = ElementPath(path)
            assert elem.tag == Tag(0x300A, 0x00B0)
            assert elem.keyword == "BeamSequence"
            assert elem.VR == "SQ"
            assert elem.item_nr == 0
            assert elem.is_sequence is True
            assert elem.child is None
            assert elem.parent is None
            assert elem.value == ""

        elem = ElementPath("BeamSequence[0].BeamName=")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 0
        assert elem.is_sequence is True
        assert elem.parent is None
        assert elem.value == ""

        child = elem.child
        assert child.tag == Tag(0x300A, 0x00C2)
        assert child.keyword == "BeamName"
        assert child.VR == "LO"
        assert child.item_nr is None
        assert child.is_sequence is False
        assert child.parent is elem
        assert child.value == ""

        elem = ElementPath("BeamSequence[0].BeamName=Test")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 0
        assert elem.is_sequence is True
        assert elem.parent is None
        assert elem.value == "Test"

        child = elem.child
        assert child.tag == Tag(0x300A, 0x00C2)
        assert child.keyword == "BeamName"
        assert child.VR == "LO"
        assert child.item_nr is None
        assert child.is_sequence is False
        assert child.parent is elem
        assert child.value == "Test"

        elem = ElementPath("BeamSequence[5].BeamSequence[13].BeamName=Test")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 5
        assert elem.is_sequence is True
        assert elem.parent is None
        assert elem.value == "Test"

        child = elem.child
        assert child.tag == Tag(0x300A, 0x00B0)
        assert child.keyword == "BeamSequence"
        assert child.VR == "SQ"
        assert child.item_nr == 13
        assert child.is_sequence is True
        assert child.parent is elem
        assert child.value == "Test"

        subchild = child.child
        assert subchild.tag == Tag(0x300A, 0x00C2)
        assert subchild.keyword == "BeamName"
        assert subchild.VR == "LO"
        assert subchild.item_nr is None
        assert subchild.is_sequence is False
        assert subchild.parent is child
        assert subchild.value == "Test"

    def test_non_sequence_repeater(self):
        """Test ElementPath using a non-sequence component."""
        elem = ElementPath("(0020,3100)")
        assert elem.tag == Tag(0x0020, 0x3100)
        assert elem.keyword == "SourceImageIDs"
        assert elem.VR == "CS"
        assert elem.item_nr is None
        assert elem.is_sequence is False

        elem = ElementPath("7Ffe,0040")
        assert elem.tag == Tag(0x7FFE, 0x0040)
        assert elem.keyword == "VariableCoefficientsSDDN"
        assert elem.VR == "OW"
        assert elem.item_nr is None
        assert elem.is_sequence is False

    def test_private(self):
        """Test using private non-sequence component."""
        elem = ElementPath("(0029,0100)")
        assert elem.tag == Tag(0x0029, 0x0100)
        assert elem.keyword == "Unknown"
        assert elem.VR == "UN"
        assert elem.item_nr is None
        assert elem.is_sequence is False

        msg = r"Unable to parse element path component: 'UnknownPrivateElement'"
        with pytest.raises(ValueError, match=msg):
            ElementPath("UnknownPrivateElement")

    def test_sequence(self):
        """Test ElementPath using a sequence component."""
        elem = ElementPath("(300a,00b0)[13]")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 13
        assert elem.is_sequence is True
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

        elem = ElementPath("300a,00b0[13]")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 13
        assert elem.is_sequence is True
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

        elem = ElementPath("BeamSequence[13]")
        assert elem.tag == Tag(0x300A, 0x00B0)
        assert elem.keyword == "BeamSequence"
        assert elem.VR == "SQ"
        assert elem.item_nr == 13
        assert elem.is_sequence is True
        assert elem.child is None
        assert elem.parent is None
        assert elem.value == ""

    def test_sequence_raises(self):
        """Test ElementPath using bad sequence component raises."""
        msg = r"Element path contains an invalid component:"
        with pytest.raises(ValueError, match=msg):
            ElementPath("(300a,00b0)[")

        with pytest.raises(ValueError, match=msg):
            ElementPath("(300a,00b0)]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("(300a,00b0)[]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("300a,00b0[")

        with pytest.raises(ValueError, match=msg):
            ElementPath("300a,00b0]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("300a,00b0[]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("BeamSequence[")

        with pytest.raises(ValueError, match=msg):
            ElementPath("BeamSequence]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("BeamSequence[]")

        with pytest.raises(ValueError, match=msg):
            ElementPath("BeamSequence[-1]")

    def test_tag(self):
        """Tests for ElementPath.tag."""
        elem = ElementPath("0000,0000")
        assert elem.tag == Tag(0x0000, 0x0000)
        elem = ElementPath("(0000,0000")
        assert elem.tag == Tag(0x0000, 0x0000)
        elem = ElementPath("0000,0000)")
        assert elem.tag == Tag(0x0000, 0x0000)
        elem = ElementPath("00(00,0000)")
        assert elem.tag == Tag(0x0000, 0x0000)
        elem = ElementPath("(00(00,00)00)")
        assert elem.tag == Tag(0x0000, 0x0000)

        # Private elements
        elem = ElementPath("(0009,0010)")
        assert elem.tag == Tag(0x0009, 0x0010)

        # Keywords
        elem = ElementPath("PatientName")
        assert elem.tag == Tag(0x0010, 0x0010)
        elem = ElementPath("CommandGroupLength")
        assert elem.tag == Tag(0x0000, 0x0000)

        # Repeater
        msg = r"Repeating group elements must be specified using " r"\(gggg,eeee\)"
        with pytest.raises(ValueError, match=msg):
            elem = ElementPath("SourceImageIDs")

        # Unknown
        msg = r"Unable to parse element path component:"
        with pytest.raises(ValueError, match=msg):
            elem = ElementPath("abcdefgh")

    def test_update_new(self):
        """Tests for parsing non-sequence strings."""
        ds = ElementPath("BeamSequence[0].PatientName=*").update(Dataset())

        assert ds.BeamSequence[0].PatientName == "*"

        paths = [
            "PatientName=Test^Name",
            "BeamSequence[0].BeamSequence[0].BeamNumber=1",
            "BeamSequence[0].BeamSequence[1].BeamNumber=2",
            "BeamSequence[0].BeamSequence[2].BeamNumber=3",
            "BeamSequence[1].BeamSequence[0].BeamNumber=4",
            "BeamSequence[1].BeamSequence[1].BeamNumber=5",
            "PatientName=Test^Name^2",
        ]
        ds = Dataset()
        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.PatientName = "Test^Name^2"
        ref.BeamSequence = [Dataset(), Dataset()]
        ref.BeamSequence[0].BeamSequence = [Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[0].BeamSequence[0].BeamNumber = 1
        ref.BeamSequence[0].BeamSequence[1].BeamNumber = 2
        ref.BeamSequence[0].BeamSequence[2].BeamNumber = 3
        ref.BeamSequence[1].BeamSequence = [Dataset(), Dataset()]
        ref.BeamSequence[1].BeamSequence[0].BeamNumber = 4
        ref.BeamSequence[1].BeamSequence[1].BeamNumber = 5

        assert ref == ds

    def test_update_repeater(self):
        """Test updating using repeater elements."""
        paths = [
            "PatientName=Test^Name",
            "BeamSequence[0].(0020,3100)=2",
        ]
        ds = Dataset()
        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.PatientName = "Test^Name"
        ref.BeamSequence = [Dataset()]
        ref.BeamSequence[0].add_new(0x00203100, "CS", "2")

        assert ref == ds

    def test_update_private(self):
        """Test updating using private elements."""
        paths = [
            "PatientName=Test^Name",
            "BeamSequence[0].(0029,0100)=2",
        ]
        ds = Dataset()
        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.PatientName = "Test^Name"
        ref.BeamSequence = [Dataset()]
        ref.BeamSequence[0].add_new(0x00290100, "UN", "2")

        assert ref == ds

    def test_update_non_sequential(self):
        """Test that updating new dataset with non-sequential items."""
        paths = [
            "BeamSequence[3]=",
            "BeamSequence[0].DACSequence[2]=",
            "BeamSequence[0].DACSequence[0].BeamSequence[1]=",
            "BeamSequence[0].DACSequence[1].BeamName=23",
        ]

        ds = Dataset()
        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.BeamSequence = [Dataset(), Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence = [Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence[0].BeamSequence = [Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence[1].BeamName = "23"

        assert ref == ds

    def test_update_existing(self):
        """Tests for parsing non-sequence strings."""
        paths = [
            "PatientName=Test^Name",
            "BeamSequence[0].DACSequence[0].BeamNumber=1",
            "BeamSequence[0].DACSequence[1].BeamNumber=2",
            "BeamSequence[0].DACSequence[2].BeamNumber=3",
            "BeamSequence[1].DACSequence[0].BeamNumber=4",
            "BeamSequence[1].DACSequence[1].BeamNumber=5",
        ]
        ds = Dataset()
        ds.PatientName = "Test^Name^2"
        ds.BeamSequence = [Dataset()]
        ds.BeamSequence[0].DACSequence = [Dataset()]
        ds.BeamSequence[0].DACSequence[0].BeamNumber = "100"

        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.PatientName = "Test^Name"
        ref.BeamSequence = [Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence = [Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence[0].BeamNumber = 1
        ref.BeamSequence[0].DACSequence[1].BeamNumber = 2
        ref.BeamSequence[0].DACSequence[2].BeamNumber = 3
        ref.BeamSequence[1].DACSequence = [Dataset(), Dataset()]
        ref.BeamSequence[1].DACSequence[0].BeamNumber = 4
        ref.BeamSequence[1].DACSequence[1].BeamNumber = 5

        assert ref == ds

    def test_update_existing_non_sequential(self):
        """Test that updating existing dataset with non-sequential items."""
        paths = [
            "PatientName=Test^Name",
            "BeamSequence[1].DACSequence[3]=",
            "BeamSequence[0].DACSequence[2]=",
            "BeamSequence[0].DACSequence[0].BeamNumber=1",
            "BeamSequence[0].DACSequence[2].BeamNumber=3",
            "BeamSequence[1].DACSequence[3].BeamNumber=4",
        ]
        ds = Dataset()
        ds.PatientName = "Test^Name^2"
        ds.BeamSequence = [Dataset(), Dataset()]
        ds.BeamSequence[0].DACSequence = [Dataset()]
        ds.BeamSequence[0].DACSequence[0].BeamNumber = "100"
        ds.BeamSequence[1].DACSequence = [Dataset(), Dataset()]
        ds.BeamSequence[1].DACSequence[1].BeamNumber = "101"
        ds.BeamSequence[1].DACSequence[0].BeamName = "Beam"

        for elem in [ElementPath(pp) for pp in paths]:
            elem.update(ds)

        ref = Dataset()
        ref.PatientName = "Test^Name"
        ref.BeamSequence = [Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence = [Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[0].DACSequence[0].BeamNumber = 1
        ref.BeamSequence[0].DACSequence[2].BeamNumber = 3
        ref.BeamSequence[1].DACSequence = [Dataset(), Dataset(), Dataset(), Dataset()]
        ref.BeamSequence[1].DACSequence[3].BeamNumber = 4
        ref.BeamSequence[1].DACSequence[0].BeamName = "Beam"
        ref.BeamSequence[1].DACSequence[1].BeamNumber = "101"

        assert ref == ds

    def test_str_types_empty(self):
        """Test that empty element values get converted correctly."""
        keywords = {
            "AE": "Receiver",
            "AS": "PatientAge",
            "AT": "OffendingElement",  # VM 1-n
            "CS": "QualityControlSubject",
            "DA": "PatientBirthDate",
            "DS": "PatientWeight",
            "DT": "AcquisitionDateTime",
            "IS": "BeamNumber",
            "LO": "DataSetSubtype",
            "LT": "ExtendedCodeMeaning",
            "PN": "PatientName",
            "SH": "CodeValue",
            "ST": "InstitutionAddress",
            "TM": "StudyTime",
            "UC": "LongCodeValue",
            "UI": "SOPClassUID",
            "UR": "CodingSchemeURL",
            "UT": "StrainAdditionalInformation",
        }
        for vr, kw in keywords.items():
            ds = ElementPath(f"{kw}=").update(Dataset())
            assert getattr(ds, kw) == ""

    def test_str_types(self):
        """Test that non-empty element values get converted correctly."""
        keywords = {
            "AE": ("Receiver", "SOME_AET"),
            "AS": ("PatientAge", "14"),
            "AT": ("OffendingElement", "00100020"),  # VM 1-n
            "CS": ("QualityControlSubject", "ASD123"),
            "DA": ("PatientBirthDate", "20000101"),
            "DS": ("PatientWeight", "14.7"),
            "DT": ("AcquisitionDateTime", "20000101120000.123456"),
            "IS": ("BeamNumber", "46"),
            "LO": ("DataSetSubtype", "Some long string thing for this one"),
            "LT": ("ExtendedCodeMeaning", "Another long string"),
            "PN": ("PatientName", "CITIZEN^Jan^X"),
            "SH": ("CodeValue", "16 character str"),
            "ST": ("InstitutionAddress", "1024 character string"),
            "TM": ("StudyTime", "120000.123456"),
            "UC": ("LongCodeValue", "So many characters in this one"),
            "UI": ("SOPClassUID", "1.2.3.4.5.6"),
            "UR": ("CodingSchemeURL", "https://github.com/pydicom/pynetdicom"),
            "UT": ("StrainAdditionalInformation", "Wheeeeeeeeee"),
        }
        for vr, (kw, val) in keywords.items():
            epath = "{}={}".format(kw, val)
            ds = ElementPath(epath).update(Dataset())
            if vr in ["DS", "IS"]:
                assert getattr(ds, kw).original_string == val
            elif vr in ["AT"]:
                assert getattr(ds, kw) == Tag(0x0010, 0x0020)
            else:
                assert getattr(ds, kw) == val

    def test_str_types_multi(self):
        """Test element values with VM > 1."""
        # AE
        epath = "RetrieveAETitle=AET1\\AET2\\AET3"
        ds = ElementPath(epath).update(Dataset())
        assert len(ds.RetrieveAETitle) == 3
        assert ds.RetrieveAETitle[0] == "AET1"
        assert ds.RetrieveAETitle[1] == "AET2"
        assert ds.RetrieveAETitle[2] == "AET3"
        # AT
        epath = "OffendingElement=00100020\\00100021\\00200020"
        ds = ElementPath(epath).update(Dataset())
        elem = ds.OffendingElement
        assert elem[0] == Tag(0x0010, 0x0020)
        assert elem[1] == Tag(0x0010, 0x0021)
        assert elem[2] == Tag(0x0020, 0x0020)
        assert len(elem) == 3

    def test_int_types_empty(self):
        """Test that empty element values get converted correctly."""
        keywords = {
            "SL": "RationalNumeratorValue",
            "SS": "SelectorSSValue",
            "UL": "SimpleFrameList",
            "US": "SourceAcquisitionBeamNumber",
        }
        for vr, kw in keywords.items():
            ds = ElementPath("{}=".format(kw)).update(Dataset())
            assert getattr(ds, kw) == ""

    def test_int_types(self):
        """Test that non-empty element values get converted correctly."""
        keywords = {
            "SL": ("RationalNumeratorValue", "-12"),
            "SS": ("SelectorSSValue", "-13"),
            "UL": ("SimpleFrameList", "14"),
            "US": ("SourceAcquisitionBeamNumber", "15"),
        }
        for vr, (kw, val) in keywords.items():
            ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
            assert getattr(ds, kw) == int(val)

    def test_int_types_multi(self):
        """Test element values with VM > 1."""
        keywords = {
            "SL": ("RationalNumeratorValue", "-12\\-13\\15"),
            "SS": ("SelectorSSValue", "-13\\-14\\16"),
            "UL": ("SimpleFrameList", "14\\15\\16"),
            "US": ("SourceAcquisitionBeamNumber", "15\\16\\17"),
        }
        kw, val = keywords["SL"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.RationalNumeratorValue == [-12, -13, 15]

        kw, val = keywords["SS"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.SelectorSSValue == [-13, -14, 16]

        kw, val = keywords["UL"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.SimpleFrameList == [14, 15, 16]

        kw, val = keywords["US"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.SourceAcquisitionBeamNumber == [15, 16, 17]

    def test_float_types_empty(self):
        """Test that empty element values get converted correctly."""
        keywords = {
            "FD": "RealWorldValueLUTData",
            "FL": "VectorAccuracy",
        }
        for vr, kw in keywords.items():
            ds = ElementPath("{}=".format(kw)).update(Dataset())
            assert getattr(ds, kw) == ""

    def test_float_types(self):
        """Test that non-empty element values get converted correctly."""
        keywords = {
            "FD": ("RealWorldValueLUTData", "-1.000005"),
            "FL": ("VectorAccuracy", "12.111102"),
        }
        for vr, (kw, val) in keywords.items():
            ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
            assert getattr(ds, kw) == float(val)

    def test_float_types_multi(self):
        """Test element values with VM > 1."""
        keywords = {
            "FD": ("RealWorldValueLUTData", "-1.000005\\91.992"),
            "FL": ("VectorAccuracy", "12.111102\\-11129.22"),
        }
        kw, val = keywords["FD"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.RealWorldValueLUTData == [-1.000005, 91.992]

        kw, val = keywords["FL"]
        ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
        assert ds.VectorAccuracy == [12.111102, -11129.22]

    def test_byte_types_empty(self):
        """Test that empty element values get converted correctly."""
        keywords = {
            "OB": "FillPattern",
            "OD": "DoubleFloatPixelData",
            "OF": "UValueData",
            "OL": "TrackPointIndexList",
            "OW": "TrianglePointIndexList",
            #'OV' : '',
            "UN": "SelectorUNValue",
        }
        for vr, kw in keywords.items():
            ds = ElementPath("{}=".format(kw)).update(Dataset())
            assert getattr(ds, kw) == ""

    def test_byte_types(self):
        """Test that non-empty element values get converted correctly."""
        keywords = {
            "OB": ("FillPattern", "00fff0ec"),
            "OD": ("DoubleFloatPixelData", "00fff0ec"),
            "OF": ("UValueData", "00fff0ec"),
            "OL": ("TrackPointIndexList", "00fff0ec"),
            "OW": ("TrianglePointIndexList", "00fff0ec"),
            #'OV' : '',''),
            "UN": ("SelectorUNValue", "00fff0ec"),
        }
        for vr, (kw, val) in keywords.items():
            ds = ElementPath("{}={}".format(kw, val)).update(Dataset())
            assert getattr(ds, kw) == b"\x00\xff\xf0\xec"


REFERENCE_FS = [
    "/test.dcm",
    "/test.txt",
    "/A/test.dcm",
    "/A/test.txt",
    "/A/B/test.dcm",
    "/A/B/test.txt",
    "/A/B/C/test.dcm",
    "/A/B/C/test.txt",
    "/A/C/test.dcm",
    "/A/C/test.txt",
]
REFERENCE_OUTPUT = [
    (["/"], False, ["/test.dcm", "/test.txt"]),
    (
        ["/"],
        True,
        [
            "/test.dcm",
            "/test.txt",
            "/A/test.dcm",
            "/A/test.txt",
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
        ],
    ),
    (["/A"], False, ["/A/test.dcm", "/A/test.txt"]),
    (
        ["/A"],
        True,
        [
            "/A/test.dcm",
            "/A/test.txt",
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
        ],
    ),
    (["/A/B"], False, ["/A/B/test.dcm", "/A/B/test.txt"]),
    (
        ["/A/B"],
        True,
        [
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
        ],
    ),
    (["/A/B/C"], False, ["/A/B/C/test.dcm", "/A/B/C/test.txt"]),
    (
        ["/A/B/C"],
        True,
        [
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
        ],
    ),
    (["/A/B/C/test.dcm"], False, ["/A/B/C/test.dcm"]),
    (["/A/B/C/test.dcm"], True, ["/A/B/C/test.dcm"]),
    # Multiples
    (["/", "/A/test.dcm"], False, ["/test.dcm", "/test.txt", "/A/test.dcm"]),
    (
        ["/", "/A/test.dcm"],
        True,
        [
            "/test.dcm",
            "/test.txt",
            "/A/test.dcm",
            "/A/test.txt",
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
        ],
    ),
    (
        ["/A/B", "/A/C"],
        False,
        ["/A/B/test.dcm", "/A/B/test.txt", "/A/C/test.dcm", "/A/C/test.txt"],
    ),
    (
        ["/A/B", "/A/C"],
        True,
        [
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
        ],
    ),
    (
        ["/A/B", "/A/C", "test.txt"],
        False,
        [
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
            "test.txt",
        ],
    ),
    (
        ["/A/B", "/A/C"],
        True,
        [
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
        ],
    ),
    (
        ["/A/B", "/A/C", "test.txt"],
        True,
        [
            "/A/B/test.dcm",
            "/A/B/test.txt",
            "/A/C/test.dcm",
            "/A/C/test.txt",
            "/A/B/C/test.dcm",
            "/A/B/C/test.txt",
            "test.txt",
        ],
    ),
]


@pytest.mark.skipif(not HAVE_PYFAKEFS, reason="Requires pyfakefs")
@pytest.mark.parametrize("fpaths, recurse, out", REFERENCE_OUTPUT)
def test_get_files(fpaths, recurse, out, fs):
    """Test finding files in a given path."""
    for fpath in REFERENCE_FS:
        fs.create_file(fpath)

    assert set(out) == set(get_files(fpaths, recurse)[0])
